"""Version control system management."""

import os
import re
import subprocess
from dataclasses import dataclass
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING, Any, ClassVar, Dict, List, MutableMapping, Optional, Type, Union

from bumpversion.ui import get_indented_logger
from bumpversion.utils import extract_regex_flags, format_and_raise_error, run_command

if TYPE_CHECKING:  # pragma: no-coverage
    from bumpversion.config import Config

from bumpversion.exceptions import DirtyWorkingDirectoryError, SignedTagsError

logger = get_indented_logger(__name__)


@dataclass
class SCMInfo:
    """Information about the current source code manager and state."""

    tool: Optional[Type["SourceCodeManager"]] = None
    commit_sha: Optional[str] = None
    distance_to_latest_tag: int = 0
    current_version: Optional[str] = None
    current_tag: Optional[str] = None
    branch_name: Optional[str] = None
    short_branch_name: Optional[str] = None
    repository_root: Optional[Path] = None
    dirty: Optional[bool] = None

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        tool_name = self.tool.__name__ if self.tool else "No SCM tool"
        return (
            f"SCMInfo(tool={tool_name}, "
            f"commit_sha={self.commit_sha}, "
            f"distance_to_latest_tag={self.distance_to_latest_tag}, "
            f"current_version={self.current_version}, "
            f"current_tag={self.current_tag}, "
            f"branch_name={self.branch_name}, "
            f"short_branch_name={self.short_branch_name}, "
            f"repository_root={self.repository_root}, "
            f"dirty={self.dirty})"
        )

    def path_in_repo(self, path: Union[Path, str]) -> bool:
        """Return whether a path is inside this repository."""
        if self.repository_root is None:
            return True
        elif not Path(path).is_absolute():
            return True

        return str(path).startswith(str(self.repository_root))


class SourceCodeManager:
    """Base class for version control systems."""

    _TEST_USABLE_COMMAND: ClassVar[List[str]] = []
    _COMMIT_COMMAND: ClassVar[List[str]] = []
    _ALL_TAGS_COMMAND: ClassVar[List[str]] = []

    @classmethod
    def commit(cls, message: str, current_version: str, new_version: str, extra_args: Optional[list] = None) -> None:
        """Commit the changes."""
        extra_args = extra_args or []
        if not current_version:
            logger.warning("No current version given, using an empty string.")
            current_version = ""
        if not new_version:
            logger.warning("No new version given, using an empty string.")
            new_version = ""

        with NamedTemporaryFile("wb", delete=False) as f:
            f.write(message.encode("utf-8"))

        env = os.environ.copy()
        env["HGENCODING"] = "utf-8"
        env["BUMPVERSION_CURRENT_VERSION"] = current_version
        env["BUMPVERSION_NEW_VERSION"] = new_version

        try:
            cmd = [*cls._COMMIT_COMMAND, f.name, *extra_args]
            run_command(cmd, env=env)
        except (subprocess.CalledProcessError, TypeError) as exc:  # pragma: no-coverage
            cls.format_and_raise_error(exc)
        finally:
            os.unlink(f.name)

    @classmethod
    def format_and_raise_error(cls, exc: Union[TypeError, subprocess.CalledProcessError]) -> None:
        """Format the error message from an exception and re-raise it as a BumpVersionError."""
        format_and_raise_error(exc)

    @classmethod
    def is_usable(cls) -> bool:
        """Is the VCS implementation usable."""
        try:
            result = run_command(cls._TEST_USABLE_COMMAND)
            return result.returncode == 0
        except (FileNotFoundError, PermissionError, NotADirectoryError, subprocess.CalledProcessError):
            return False

    @classmethod
    def assert_nondirty(cls) -> None:
        """Assert that the working directory is not dirty."""
        raise NotImplementedError()

    @classmethod
    def latest_tag_info(cls, tag_name: str, parse_pattern: str) -> SCMInfo:
        """Return information about the latest tag."""
        raise NotImplementedError()

    @classmethod
    def add_path(cls, path: Union[str, Path]) -> None:
        """Add a path to the VCS."""
        raise NotImplementedError()

    @classmethod
    def tag(cls, name: str, sign: bool = False, message: Optional[str] = None) -> None:
        """Create a tag of the new_version in VCS."""
        raise NotImplementedError

    @classmethod
    def get_all_tags(cls) -> List[str]:
        """Return all tags in VCS."""
        try:
            result = run_command(cls._ALL_TAGS_COMMAND)
            return result.stdout.splitlines()
        except (FileNotFoundError, PermissionError, NotADirectoryError, subprocess.CalledProcessError):
            return []

    @classmethod
    def get_version_from_tag(cls, tag: str, tag_name: str, parse_pattern: str) -> Optional[str]:
        """Return the version from a tag."""
        version_pattern = parse_pattern.replace("\\\\", "\\")
        version_pattern, regex_flags = extract_regex_flags(version_pattern)
        parts = tag_name.split("{new_version}", maxsplit=1)
        prefix = parts[0]
        suffix = parts[1]
        rep = f"{regex_flags}{re.escape(prefix)}(?P<current_version>{version_pattern}){re.escape(suffix)}"
        tag_regex = re.compile(rep)
        return match["current_version"] if (match := tag_regex.search(tag)) else None

    @classmethod
    def commit_to_scm(
        cls,
        files: List[Union[str, Path]],
        config: "Config",
        context: MutableMapping,
        extra_args: Optional[List[str]] = None,
        dry_run: bool = False,
    ) -> None:
        """Commit the files to the source code management system."""
        if not cls.is_usable():
            logger.error("SCM tool '%s' is unusable, unable to commit.", cls.__name__)
            return

        if not config.commit:
            logger.info("Would not commit")
            return

        do_commit = not dry_run
        logger.info(
            "%s %s commit",
            "Preparing" if do_commit else "Would prepare",
            cls.__name__,
        )
        logger.indent()
        for path in files:
            logger.info(
                "%s changes in file '%s' to %s",
                "Adding" if do_commit else "Would add",
                path,
                cls.__name__,
            )

            if do_commit:
                cls.add_path(path)

        commit_message = config.message.format(**context)

        logger.info(
            "%s to %s with message '%s'",
            "Committing" if do_commit else "Would commit",
            cls.__name__,
            commit_message,
        )
        if do_commit:
            cls.commit(
                message=commit_message,
                current_version=context["current_version"],
                new_version=context["new_version"],
                extra_args=extra_args,
            )
        logger.dedent()

    @classmethod
    def tag_in_scm(cls, config: "Config", context: MutableMapping, dry_run: bool = False) -> None:
        """Tag the current commit in the source code management system."""
        if not config.tag:
            logger.info("Would not tag")
            return
        sign_tags = config.sign_tags
        tag_name = config.tag_name.format(**context)
        tag_message = config.tag_message.format(**context)
        existing_tags = cls.get_all_tags()
        do_tag = not dry_run

        if tag_name in existing_tags:
            logger.warning("Tag '%s' already exists. Will not tag.", tag_name)
            return

        logger.info(
            "%s '%s' %s in %s and %s",
            "Tagging" if do_tag else "Would tag",
            tag_name,
            f"with message '{tag_message}'" if tag_message else "without message",
            cls.__name__,
            "signing" if sign_tags else "not signing",
        )
        if do_tag:
            cls.tag(tag_name, sign_tags, tag_message)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f"{self.__class__.__name__}"


class Git(SourceCodeManager):
    """Git implementation."""

    _TEST_USABLE_COMMAND: ClassVar[List[str]] = ["git", "rev-parse", "--git-dir"]
    _COMMIT_COMMAND: ClassVar[List[str]] = ["git", "commit", "-F"]
    _ALL_TAGS_COMMAND: ClassVar[List[str]] = ["git", "tag", "--list"]

    @classmethod
    def assert_nondirty(cls) -> None:
        """Assert that the working directory is not dirty."""
        lines = [
            line.strip()
            for line in run_command(["git", "status", "--porcelain"]).stdout.splitlines()
            if not line.strip().startswith("??")
        ]
        if joined_lines := "\n".join(lines):
            raise DirtyWorkingDirectoryError(f"Git working directory is not clean:\n\n{joined_lines}")

    @classmethod
    def latest_tag_info(cls, tag_name: str, parse_pattern: str) -> SCMInfo:
        """Return information about the latest tag."""
        if not cls.is_usable():
            return SCMInfo()

        info: Dict[str, Any] = {"tool": cls}

        try:
            # git-describe doesn't update the git-index, so we do that
            run_command(["git", "update-index", "--refresh", "-q"])
        except subprocess.CalledProcessError as e:
            logger.debug("Error when running git update-index: %s", e.stderr)

        commit_info = cls._commit_info(parse_pattern, tag_name)
        rev_info = cls._revision_info()
        info.update(commit_info)
        info.update(rev_info)

        return SCMInfo(**info)

    @classmethod
    def _commit_info(cls, parse_pattern: str, tag_name: str) -> dict:
        """
        Get the commit info for the repo.

        Args:
            parse_pattern: The regular expression pattern used to parse the version from the tag.
            tag_name: The tag name format used to locate the latest tag.

        Returns:
            A dictionary containing information about the latest commit.
        """
        tag_pattern = tag_name.replace("{new_version}", "*")
        info = dict.fromkeys(["dirty", "commit_sha", "distance_to_latest_tag", "current_version", "current_tag"])
        info["distance_to_latest_tag"] = 0
        try:
            # get info about the latest tag in git
            git_cmd = [
                "git",
                "describe",
                "--dirty",
                "--tags",
                "--long",
                "--abbrev=40",
                f"--match={tag_pattern}",
            ]
            result = run_command(git_cmd)
            describe_out = result.stdout.strip().split("-")
            if describe_out[-1].strip() == "dirty":
                info["dirty"] = True
                describe_out.pop()
            else:
                info["dirty"] = False

            info["commit_sha"] = describe_out.pop().lstrip("g")
            info["distance_to_latest_tag"] = int(describe_out.pop())
            info["current_tag"] = "-".join(describe_out)
            version = cls.get_version_from_tag("-".join(describe_out), tag_name, parse_pattern)
            info["current_version"] = version or "-".join(describe_out).lstrip("v")
        except subprocess.CalledProcessError as e:
            logger.debug("Error when running git describe: %s", e.stderr)

        return info

    @classmethod
    def _revision_info(cls) -> dict:
        """
        Returns a dictionary containing revision information.

        If an error occurs while running the git command, the dictionary values will be set to None.

        Returns:
            A dictionary with the following keys:
                - branch_name: The name of the current branch.
                - short_branch_name: A 20 lowercase characters of the branch name with special characters removed.
                - repository_root: The root directory of the Git repository.
        """
        info = dict.fromkeys(["branch_name", "short_branch_name", "repository_root"])

        try:
            git_cmd = ["git", "rev-parse", "--show-toplevel", "--abbrev-ref", "HEAD"]
            result = run_command(git_cmd)
            lines = [line.strip() for line in result.stdout.split("\n")]
            repository_root = Path(lines[0])
            branch_name = lines[1]
            short_branch_name = re.sub(r"([^a-zA-Z0-9]*)", "", branch_name).lower()[:20]
            info["branch_name"] = branch_name
            info["short_branch_name"] = short_branch_name
            info["repository_root"] = repository_root
        except subprocess.CalledProcessError as e:
            logger.debug("Error when running git rev-parse: %s", e.stderr)

        return info

    @classmethod
    def add_path(cls, path: Union[str, Path]) -> None:
        """Add a path to the VCS."""
        info = SCMInfo(**cls._revision_info())
        if not info.path_in_repo(path):
            return
        cwd = Path.cwd()
        temp_path = os.path.relpath(path, cwd)
        try:
            run_command(["git", "add", "--update", str(temp_path)])
        except subprocess.CalledProcessError as e:
            format_and_raise_error(e)

    @classmethod
    def tag(cls, name: str, sign: bool = False, message: Optional[str] = None) -> None:
        """
        Create a tag of the new_version in VCS.

        If only name is given, bumpversion uses a lightweight tag.
        Otherwise, it uses an annotated tag.

        Args:
            name: The name of the tag
            sign: True to sign the tag
            message: An optional message to annotate the tag.
        """
        command = ["git", "tag", name]
        if sign:
            command += ["--sign"]
        if message:
            command += ["--message", message]
        run_command(command)


class Mercurial(SourceCodeManager):
    """Mercurial implementation."""

    _TEST_USABLE_COMMAND: ClassVar[List[str]] = ["hg", "root"]
    _COMMIT_COMMAND: ClassVar[List[str]] = ["hg", "commit", "--logfile"]
    _ALL_TAGS_COMMAND: ClassVar[List[str]] = ["hg", "log", '--rev="tag()"', '--template="{tags}\n"']

    @classmethod
    def latest_tag_info(cls, tag_name: str, parse_pattern: str) -> SCMInfo:
        """Return information about the latest tag."""
        current_version = None
        re_pattern = tag_name.replace("{new_version}", ".*")
        result = run_command(["hg", "log", "-r", f"tag('re:{re_pattern}')", "--template", "{latesttag}\n"])
        result.check_returncode()
        if result.stdout:
            tag_string = result.stdout.splitlines(keepends=False)[-1]
            current_version = cls.get_version_from_tag(tag_string, tag_name, parse_pattern)
        else:
            logger.debug("No tags found")
        is_dirty = len(run_command(["hg", "status", "-mard"]).stdout) != 0
        return SCMInfo(tool=cls, current_version=current_version, dirty=is_dirty)

    @classmethod
    def assert_nondirty(cls) -> None:
        """Assert that the working directory is clean."""
        lines = [
            line.strip()
            for line in run_command(["hg", "status", "-mard"]).stdout.splitlines()
            if not line.strip().startswith("??")
        ]

        if lines:
            joined_lines = "\n".join(lines)
            raise DirtyWorkingDirectoryError(f"Mercurial working directory is not clean:\n{joined_lines}")

    @classmethod
    def add_path(cls, path: Union[str, Path]) -> None:
        """Add a path to the VCS."""
        pass

    @classmethod
    def tag(cls, name: str, sign: bool = False, message: Optional[str] = None) -> None:
        """
        Create a tag of the new_version in VCS.

        If only name is given, bumpversion uses a lightweight tag.
        Otherwise, it uses an annotated tag.

        Args:
            name: The name of the tag
            sign: True to sign the tag
            message: A optional message to annotate the tag.

        Raises:
            SignedTagsError: If ``sign`` is ``True``
        """
        command = ["hg", "tag", name]
        if sign:
            raise SignedTagsError("Mercurial does not support signed tags.")
        if message:
            command += ["--message", message]
        run_command(command)


def get_scm_info(tag_name: str, parse_pattern: str) -> SCMInfo:
    """Return a dict with the latest source code management info."""
    if Git.is_usable():
        return Git.latest_tag_info(tag_name, parse_pattern)
    elif Mercurial.is_usable():
        return Mercurial.latest_tag_info(tag_name, parse_pattern)
    else:
        return SCMInfo()



"""Tests for the bump module."""

from pathlib import Path
import shutil
from textwrap import dedent
from unittest.mock import MagicMock, patch

import pytest

from bumpversion import bump
from bumpversion.exceptions import ConfigurationError, VersionNotFoundError
from bumpversion.files import ConfiguredFile
from bumpversion.scm import Git, SCMInfo
from bumpversion.utils import run_command
from tests.conftest import get_config_data, inside_dir


@pytest.fixture
def mock_context():
    return {"current_version": "1.2.3", "new_version": "1.2.4"}


class TestGetNextVersion:
    """Tests for the get_next_version function."""

    def test_passing_a_new_version_should_override_a_version_part(self, tmp_path: Path):
        """Passing a new version should return that version."""
        # Arrange
        with inside_dir(tmp_path):
            config, version_config, current_version = get_config_data({"current_version": "0.1.0"})
            version_part = "patch"
            new_version = "1.2.3"
            expected_next_version = version_config.parse("1.2.3")

            # Act
            actual_next_version = bump.get_next_version(current_version, config, version_part, new_version)

        # Assert
        assert actual_next_version == expected_next_version

    def test_passing_version_part_should_increment_part(self):
        """If a version part is provided, it should return the increment of that part."""
        # Arrange
        config, version_config, current_version = get_config_data({"current_version": "0.1.0"})
        version_part = "major"
        new_version = None
        expected_next_version = version_config.parse("1.0.0")

        # Act
        actual_next_version = bump.get_next_version(current_version, config, version_part, new_version)

        # Assert
        assert actual_next_version == expected_next_version

    def test_passing_no_arguments_raises_error(self):
        # Arrange
        current_version = MagicMock()
        config = MagicMock()
        version_part = None
        new_version = None

        # Act/Assert
        with pytest.raises(ConfigurationError):
            bump.get_next_version(current_version, config, version_part, new_version)


class TestDoBump:
    """Tests for the do_bump function."""

    @patch("bumpversion.files.modify_files")
    @patch("bumpversion.bump.update_config_file")
    def test_passing_version_part_increments_part(self, mock_update_config_file, mock_modify_files):
        # Arrange
        version_part = "major"
        new_version = None
        config, version_config, current_version = get_config_data(
            {
                "current_version": "1.2.3",
                "files": [{"filename": "foo.txt"}, {"filename": "bar.txt"}],
            }
        )
        config.scm_info = SCMInfo()
        dry_run = False

        # Act
        bump.do_bump(version_part, new_version, config, dry_run=dry_run)

        # Assert
        mock_modify_files.assert_called_once()
        mock_update_config_file.assert_called_once()
        assert {f.file_change.filename for f in mock_modify_files.call_args[0][0]} == {
            "foo.txt",
            "bar.txt",
        }
        assert mock_update_config_file.call_args[0][0] is None
        assert mock_update_config_file.call_args[0][1] == config
        assert mock_update_config_file.call_args[0][2] == current_version
        assert mock_update_config_file.call_args[0][3] == current_version.bump(version_part)
        assert mock_update_config_file.call_args[0][5] is False

    @patch("bumpversion.files.modify_files")
    @patch("bumpversion.bump.update_config_file")
    def test_passing_new_version_sets_version(self, mock_update_config_file, mock_modify_files):
        # Arrange
        version_part = None
        new_version = "2.0.0"
        config, version_config, current_version = get_config_data(
            {
                "current_version": "1.2.3",
            }
        )
        config.scm_info = SCMInfo()
        dry_run = True

        # Act
        bump.do_bump(version_part, new_version, config, dry_run=dry_run)

        # Assert
        mock_modify_files.assert_called_once()
        assert mock_modify_files.call_args[0][0] == []
        assert mock_modify_files.call_args[0][1] == current_version
        assert mock_modify_files.call_args[0][2] == version_config.parse(new_version)

        mock_update_config_file.assert_called_once()
        assert mock_update_config_file.call_args[0][0] is None
        assert mock_update_config_file.call_args[0][1] == config
        assert mock_update_config_file.call_args[0][2] == current_version
        assert mock_update_config_file.call_args[0][3] == version_config.parse(new_version)
        assert mock_update_config_file.call_args[0][5] is dry_run

    @patch("bumpversion.bump.commit_and_tag")
    @patch("bumpversion.bump.update_config_file")
    def test_doesnt_commit_if_modify_error(
        self, mock_update_config_file, mock_commit_and_tag, tmp_path: Path, fixtures_path: Path
    ):
        from bumpversion import config

        # Arrange
        setup_py_path = tmp_path / "setup.py"
        setup_py_path.touch()
        init_path = tmp_path / "bumpversion/__init__.py"
        init_path.parent.mkdir(parents=True)
        init_path.touch()
        orig_config_path = fixtures_path / "basic_cfg.toml"
        dest_config_path = tmp_path / "pyproject.toml"
        shutil.copyfile(orig_config_path, dest_config_path)
        version_part = "patch"

        # Act
        with inside_dir(tmp_path):
            config = config.get_configuration(config_file=dest_config_path)
            with pytest.raises(VersionNotFoundError):
                bump.do_bump(version_part, None, config)

        # Assert
        mock_commit_and_tag.assert_not_called()

        mock_update_config_file.assert_not_called()

    @patch("bumpversion.files.modify_files")
    @patch("bumpversion.bump.update_config_file")
    def test_when_new_equals_current_nothing_happens(self, mock_update_config_file, mock_modify_files, tmp_path: Path):
        """When the new version is the same as the current version, nothing should happen."""

        # Arrange
        version_part = None
        new_version = "1.2.3"

        with inside_dir(tmp_path):
            config, version_config, current_version = get_config_data({"current_version": "1.2.3"})
            # Act
            bump.do_bump(version_part, new_version, config)

        # Assert
        mock_modify_files.assert_not_called()
        mock_update_config_file.assert_not_called()

    def test_passing_no_arguments_raises_error(self):
        # Arrange
        version_part = None
        new_version = None
        config, version_config, current_version = get_config_data({"current_version": "1.2.3"})
        config.scm_info = SCMInfo()
        dry_run = False

        # Act/Assert
        with pytest.raises(ConfigurationError):
            bump.do_bump(version_part, new_version, config, dry_run=dry_run)

    @patch("bumpversion.files.modify_files")
    @patch("bumpversion.bump.update_config_file")
    def test_includes_files_with_include_bumps(self, mock_update_config_file, mock_modify_files):
        """Files that have include_bumps are included when those bumps happen."""
        # Arrange
        new_version = None
        config, version_config, current_version = get_config_data(
            {
                "current_version": "1.2.3",
                "files": [{"filename": "foo.txt", "include_bumps": ["major", "minor"]}, {"filename": "bar.txt"}],
            }
        )
        config.scm_info = SCMInfo()
        dry_run = False

        # Act
        bump.do_bump("patch", new_version, config, dry_run=dry_run)

        # Assert
        assert {f.file_change.filename for f in mock_modify_files.call_args[0][0]} == {"bar.txt"}

        # Act
        bump.do_bump("minor", new_version, config, dry_run=dry_run)

        # Assert
        assert {f.file_change.filename for f in mock_modify_files.call_args[0][0]} == {"foo.txt", "bar.txt"}

    @patch("bumpversion.files.modify_files")
    @patch("bumpversion.bump.update_config_file")
    def test_excludes_files_with_exclude_bumps(self, mock_update_config_file, mock_modify_files):
        """Files that have exclude_bumps are excluded when those bumps happen."""
        # Arrange
        new_version = None
        config, version_config, current_version = get_config_data(
            {
                "current_version": "1.2.3",
                "files": [{"filename": "foo.txt", "exclude_bumps": ["patch"]}, {"filename": "bar.txt"}],
            }
        )
        config.scm_info = SCMInfo()
        dry_run = False

        # Act
        bump.do_bump("patch", new_version, config, dry_run=dry_run)

        # Assert
        assert {f.file_change.filename for f in mock_modify_files.call_args[0][0]} == {"bar.txt"}

        # Act
        bump.do_bump("minor", new_version, config, dry_run=dry_run)

        # Assert
        assert {f.file_change.filename for f in mock_modify_files.call_args[0][0]} == {"foo.txt", "bar.txt"}


def test_commit_and_tag_with_no_tool():
    config, version_config, current_version = get_config_data(
        {
            "current_version": "1.2.3",
        }
    )
    config.scm_info = SCMInfo()
    mock_context = MagicMock()

    bump.commit_and_tag(config, None, [], mock_context, False)


def test_commit_and_tag_with_tool(mock_context):
    config, version_config, current_version = get_config_data(
        {"current_version": "1.2.3", "files": [{"filename": "foo.txt"}, {"filename": "bar.txt"}]}
    )
    config.scm_info = SCMInfo()
    config.scm_info.tool = MagicMock(spec=Git)
    configured_files = [ConfiguredFile(file_cfg, version_config) for file_cfg in config.files]
    bump.commit_and_tag(config, None, configured_files, mock_context, False)
    config.scm_info.tool.commit_to_scm.assert_called_once()
    config.scm_info.tool.tag_in_scm.assert_called_once()
    assert set(config.scm_info.tool.commit_to_scm.call_args[0][0]) == {"foo.txt", "bar.txt"}


def test_commit_and_tag_with_config_file(mock_context):
    config, version_config, current_version = get_config_data(
        {"current_version": "1.2.3", "files": [{"filename": "foo.txt"}, {"filename": "bar.txt"}]}
    )
    config.scm_info = SCMInfo()
    config.scm_info.tool = MagicMock(spec=Git)
    configured_files = [ConfiguredFile(file_cfg, version_config) for file_cfg in config.files]
    bump.commit_and_tag(config, Path("pyproject.toml"), configured_files, mock_context, False)
    config.scm_info.tool.commit_to_scm.assert_called_once()
    config.scm_info.tool.tag_in_scm.assert_called_once()
    assert set(config.scm_info.tool.commit_to_scm.call_args[0][0]) == {"foo.txt", "bar.txt", "pyproject.toml"}


def test_key_path_required_for_toml_change(tmp_path: Path, caplog):
    """If the key_path is not provided, the toml file should use standard search and replace."""
    from bumpversion import config
    from bumpversion.versioning.version_config import VersionConfig

    # Arrange
    config_path = tmp_path / "pyproject.toml"
    config_path.write_text(
        dedent(
            """
            [project]
            version = "0.1.26"

            [tool.bumpversion]
            current_version = "0.1.26"
            allow_dirty = true
            commit = true

            [[tool.bumpversion.files]]
            filename = "pyproject.toml"
            search = "version = \\"{current_version}\\""
            replace = "version = \\"{new_version}\\""

            [tool.othertool]
            bake_cookies = true
            ignore-words-list = "sugar, salt, flour"
            """
        ),
        encoding="utf-8",
    )

    conf = config.get_configuration(config_file=config_path)

    # Act
    from click.testing import CliRunner, Result
    from bumpversion import cli

    runner: CliRunner = CliRunner()
    with inside_dir(tmp_path):
        result: Result = runner.invoke(
            cli.cli,
            ["bump", "-vv", "minor"],
        )

    if result.exit_code != 0:
        print(caplog.text)
        print("Here is the output:")
        print(result.output)
        import traceback

        print(traceback.print_exception(result.exc_info[1]))

    # Assert
    assert result.exit_code == 0
    assert config_path.read_text() == dedent(
        """
        [project]
        version = "0.2.0"

        [tool.bumpversion]
        current_version = "0.2.0"
        allow_dirty = true
        commit = true

        [[tool.bumpversion.files]]
        filename = "pyproject.toml"
        search = "version = \\"{current_version}\\""
        replace = "version = \\"{new_version}\\""

        [tool.othertool]
        bake_cookies = true
        ignore-words-list = "sugar, salt, flour"
        """
    )


def test_changes_to_files_are_committed(git_repo: Path, caplog):
    """Any files changed during the bump are committed."""
    from bumpversion import config

    # Arrange
    config_path = git_repo / ".bumpversion.toml"
    config_path.write_text(
        dedent(
            """
            [tool.bumpversion]
            current_version = "0.1.26"
            tag_name = "{new_version}"
            commit = true

            [[tool.bumpversion.files]]
            glob = "helm/charts/*/Chart.yaml"

            """
        ),
        encoding="utf-8",
    )
    chart_contents = dedent(
        """
        appVersion: 0.1.26
        version: 0.1.26

        """
    )
    chart1_path = git_repo / "helm" / "charts" / "somechart" / "Chart.yaml"
    chart2_path = git_repo / "helm" / "charts" / "otherchart" / "Chart.yaml"
    chart1_path.parent.mkdir(parents=True, exist_ok=True)
    chart1_path.write_text(chart_contents, encoding="utf-8")
    chart2_path.parent.mkdir(parents=True, exist_ok=True)
    chart2_path.write_text(chart_contents, encoding="utf-8")

    with inside_dir(git_repo):
        run_command(["git", "add", str(chart1_path), str(chart2_path), str(config_path)])
        run_command(["git", "commit", "-m", "Initial commit"])
        run_command(["git", "tag", "0.1.26"])

    # Act
    from click.testing import CliRunner, Result
    from bumpversion import cli

    runner: CliRunner = CliRunner()
    with inside_dir(git_repo):
        result: Result = runner.invoke(
            cli.cli,
            ["bump", "-vv", "minor"],
        )

    if result.exit_code != 0:
        print(caplog.text)
        print("Here is the output:")
        print(result.output)
        import traceback

        print(traceback.print_exception(result.exc_info[1]))

    # Assert
    assert result.exit_code == 0
    assert config_path.read_text() == dedent(
        """
        [tool.bumpversion]
        current_version = "0.2.0"
        tag_name = "{new_version}"
        commit = true

        [[tool.bumpversion.files]]
        glob = "helm/charts/*/Chart.yaml"

        """
    )
    assert chart1_path.read_text() == dedent(
        """
        appVersion: 0.2.0
        version: 0.2.0

        """
    )
    assert chart2_path.read_text() == dedent(
        """
        appVersion: 0.2.0
        version: 0.2.0

        """
    )
    with inside_dir(git_repo):
        status = run_command(["git", "status", "--porcelain"])
        assert status.stdout == ""



"""Tests of the VCS module."""

import logging
import shutil
import subprocess
from pathlib import Path

import pytest
from pytest import param, LogCaptureFixture

from bumpversion import scm
from bumpversion.exceptions import DirtyWorkingDirectoryError, BumpVersionError
from bumpversion.ui import setup_logging
from bumpversion.utils import run_command
from tests.conftest import get_config_data, inside_dir


class TestSourceCodeManager:
    """Tests related to SourceCodeManager."""

    class TestGetVersionFromTag:
        """Tests for the get_version_from_tag classmethod."""

        simple_pattern = r"(?P<major>\\d+)\\.(?P<minor>\\d+)\.(?P<patch>\\d+)"
        complex_pattern = r"(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)(-(?P<release>[0-9A-Za-z]+))?(\\+build\\.(?P<build>.[0-9A-Za-z]+))?"

        @pytest.mark.parametrize(
            ["tag", "tag_name", "parse_pattern", "expected"],
            [
                param("1.2.3", "{new_version}", simple_pattern, "1.2.3", id="no-prefix, no-suffix"),
                param("v/1.2.3", "v/{new_version}", simple_pattern, "1.2.3", id="prefix, no-suffix"),
                param("v/1.2.3/12345", "v/{new_version}/{something}", simple_pattern, "1.2.3", id="prefix, suffix"),
                param("1.2.3/2024-01-01", "{new_version}/{today}", simple_pattern, "1.2.3", id="no-prefix, suffix"),
                param(
                    "app/4.0.3-beta+build.1047",
                    "app/{new_version}",
                    complex_pattern,
                    "4.0.3-beta+build.1047",
                    id="complex",
                ),
            ],
        )
        def returns_version_from_pattern(self, tag: str, tag_name: str, parse_pattern: str, expected: str) -> None:
            """It properly returns the version from the tag."""
            # Act
            version = scm.SourceCodeManager.get_version_from_tag(tag, tag_name, parse_pattern)

            # Assert
            assert version == expected

    def test_format_and_raise_error_returns_scm_error(self, git_repo: Path) -> None:
        """The output formatting from called process error string includes the underlying scm error."""
        with inside_dir(git_repo):
            try:
                run_command(["git", "add", "newfile.txt"])
            except subprocess.CalledProcessError as e:
                with pytest.raises(BumpVersionError) as bump_error:
                    scm.Git.format_and_raise_error(e)
                assert bump_error.value.message == (
                    "Failed to run `git add newfile.txt`: return code 128, output: "
                    "fatal: pathspec 'newfile.txt' did not match any files\n"
                )


class TestGit:
    """Tests related to Git."""

    class TestIsUsable:
        """Tests related to is_usable."""

        def test_recognizes_a_git_repo(self, git_repo: Path) -> None:
            """Should return true if git is available, and it is a git repo."""
            with inside_dir(git_repo):
                assert scm.Git.is_usable()

        def test_recognizes_not_a_git_repo(self, tmp_path: Path) -> None:
            """Should return false if it is not a git repo."""
            with inside_dir(tmp_path):
                assert not scm.Git.is_usable()

    class TestAssertNonDirty:
        """Tests for the Git.assert_nondirty() function."""

        def test_does_nothing_when_not_dirty(self, git_repo: Path) -> None:
            """If the git repo is clean, assert_nondirty should do nothing."""
            with inside_dir(git_repo):
                scm.Git.assert_nondirty()

        def test_raises_error_when_dirty(self, git_repo: Path) -> None:
            """If the git repo has modified files, assert_nondirty should return false."""
            readme = git_repo.joinpath("readme.md")
            readme.touch()
            with pytest.raises(DirtyWorkingDirectoryError):
                with inside_dir(git_repo):
                    subprocess.run(["git", "add", "readme.md"])
                    scm.Git.assert_nondirty()

    class TestLatestTagInfo:
        """Test for the Git.latest_tag_info() function."""

        def test_an_empty_repo_only_fills_tool_info(self, git_repo: Path) -> None:
            """
            If the repo is empty, the info only includes the tool type.

            The git commands to get all the other information depend on there being a `HEAD` object.
            """
            tag_prefix = "app/"
            parse_pattern = r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)"
            tag_name = f"{tag_prefix}{{new_version}}"
            expected = scm.SCMInfo(
                tool=scm.Git,
                commit_sha=None,
                distance_to_latest_tag=0,
                current_version=None,
                branch_name=None,
                short_branch_name=None,
                repository_root=None,
                dirty=None,
            )
            with inside_dir(git_repo):
                latest_tag_info = scm.Git.latest_tag_info(tag_name, parse_pattern=parse_pattern)
                assert latest_tag_info == expected

        def test_returns_commit_and_tag_info(self, git_repo: Path) -> None:
            """Should return information that it is a git repo."""
            tag_prefix = "app/"
            parse_pattern = r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)"
            tag_name = f"{tag_prefix}{{new_version}}"
            readme = git_repo.joinpath("readme.md")
            readme.touch()
            with inside_dir(git_repo):
                # Add a file and tag
                subprocess.run(["git", "add", "readme.md"])
                subprocess.run(["git", "commit", "-m", "first"])
                subprocess.run(["git", "tag", f"{tag_prefix}0.1.0"])
                tag_info = scm.Git.latest_tag_info(tag_name, parse_pattern=parse_pattern)
                assert tag_info.commit_sha is not None
                assert tag_info.current_version == "0.1.0"
                assert tag_info.current_tag == f"{tag_prefix}0.1.0"
                assert tag_info.distance_to_latest_tag == 0
                assert tag_info.branch_name == "master"
                assert tag_info.short_branch_name == "master"
                assert tag_info.repository_root == git_repo
                assert tag_info.dirty is False


def test_git_detects_existing_tag(git_repo: Path, caplog: LogCaptureFixture) -> None:
    """Attempting to tag when a tag exists will do nothing."""
    # Arrange
    git_repo.joinpath("readme.md").touch()
    git_repo.joinpath("something.md").touch()
    overrides = {"current_version": "0.1.0", "commit": True, "tag": True}
    context = {
        "current_version": "0.1.0",
        "new_version": "0.2.0",
    }
    setup_logging(2)
    with inside_dir(git_repo):
        conf, version_config, current_version = get_config_data(overrides)
        subprocess.run(["git", "add", "readme.md"])
        subprocess.run(["git", "commit", "-m", "first"])
        subprocess.run(["git", "tag", "v0.1.0"])
        subprocess.run(["git", "add", "something.md"])
        subprocess.run(["git", "commit", "-m", "second"])
        subprocess.run(["git", "tag", "v0.2.0"])

        # Act
        scm.Git.tag_in_scm(config=conf, context=context)

    # Assert
    assert "Will not tag" in caplog.text


def test_hg_is_not_usable(tmp_path: Path) -> None:
    """Should return false if it is not a mercurial repo."""
    with inside_dir(tmp_path):
        assert not scm.Mercurial.is_usable()


@pytest.mark.skipif(not shutil.which("hg"), reason="Mercurial is not available.")
def test_hg_is_usable(hg_repo: Path) -> None:
    """Should return false if it is not a mercurial repo."""
    with inside_dir(hg_repo):
        assert scm.Mercurial.is_usable()


@pytest.mark.parametrize(
    ["repo", "scm_command", "scm_class"],
    [
        param(
            "git_repo",
            "git",
            scm.Git,
            id="git",
            marks=pytest.mark.skipif(not shutil.which("git"), reason="Git is not available."),
        ),
        param(
            "hg_repo",
            "hg",
            scm.Mercurial,
            id="hg",
            marks=pytest.mark.skipif(not shutil.which("hg"), reason="Mercurial is not available."),
        ),
    ],
)
def test_commit_and_tag_from_below_scm_root(repo: str, scm_command: str, scm_class: scm.SourceCodeManager, request):
    # Arrange
    repo_path: Path = request.getfixturevalue(repo)
    version_path = repo_path / "VERSION"
    version_path.write_text("30.0.3", encoding="utf-8")
    sub_dir_path = repo_path / "subdir"
    sub_dir_path.mkdir(exist_ok=True)

    overrides = {"current_version": "30.0.3", "commit": True, "tag": True, "files": [{"filename": str(version_path)}]}
    context = {
        "current_version": "30.0.3",
        "new_version": "30.1.0",
    }
    with inside_dir(repo_path):
        conf, version_config, current_version = get_config_data(overrides)
        subprocess.run([scm_command, "add", "VERSION"], check=True, capture_output=True)
        subprocess.run([scm_command, "commit", "-m", "initial commit"], check=True, capture_output=True)
        with inside_dir(sub_dir_path):
            version_path.write_text("30.1.0", encoding="utf-8")

            # Act
            scm_class.commit_to_scm(files=[version_path], config=conf, context=context)
            scm_class.tag_in_scm(config=conf, context=context)

        # Assert
        tag_info = scm_class.latest_tag_info(conf.tag_name, parse_pattern=conf.parse)
        if scm_command == "git":
            assert tag_info.commit_sha is not None
            assert tag_info.distance_to_latest_tag == 0
        assert tag_info.current_version == "30.1.0"
        assert tag_info.dirty is False


# write tests for no-commit, no-tag and dry-run
@pytest.mark.parametrize(
    ["repo", "scm_command", "scm_class"],
    [
        param(
            "git_repo",
            "git",
            scm.Git,
            id="git",
            marks=pytest.mark.skipif(not shutil.which("git"), reason="Git is not available."),
        ),
        param(
            "hg_repo",
            "hg",
            scm.Mercurial,
            id="hg",
            marks=pytest.mark.skipif(not shutil.which("hg"), reason="Mercurial is not available."),
        ),
    ],
)
@pytest.mark.parametrize(
    ["commit", "tag", "dry_run", "should_commit", "should_tag"],
    [
        param(True, True, True, False, False, id="dry-run-stops-commit-and-tag"),
        param(True, False, False, True, False, id="commit-no-tag"),
        param(False, True, False, False, True, id="no-commit-will-tag"),
    ],
)
def test_commit_tag_dry_run_interactions(
    repo: str,
    scm_command: str,
    scm_class: scm.SourceCodeManager,
    commit: bool,
    tag: bool,
    dry_run: bool,
    should_commit: bool,
    should_tag: bool,
    caplog: LogCaptureFixture,
    request,
):
    """Combinations of commit, tag, dry-run and should produce the expected results."""
    # Arrange
    caplog.set_level(logging.INFO)
    repo_path: Path = request.getfixturevalue(repo)
    version_path = repo_path / "VERSION"
    version_path.write_text("30.0.3", encoding="utf-8")

    overrides = {"current_version": "30.0.3", "commit": commit, "tag": tag, "files": [{"filename": str(version_path)}]}
    context = {
        "current_version": "30.0.3",
        "new_version": "30.1.0",
    }
    with inside_dir(repo_path):
        conf, version_config, current_version = get_config_data(overrides)
        subprocess.run([scm_command, "add", "VERSION"], check=True, capture_output=True)
        subprocess.run([scm_command, "commit", "-m", "initial commit"], check=True, capture_output=True)
        version_path.write_text("30.1.0", encoding="utf-8")
        # Act
        scm_class.commit_to_scm(files=[version_path], config=conf, context=context, dry_run=dry_run)
        scm_class.tag_in_scm(config=conf, context=context, dry_run=dry_run)

        # Assert
        if commit and dry_run:
            assert "Would commit" in caplog.text
        elif should_commit:
            assert "Committing " in caplog.text
        else:
            assert "Would not commit" in caplog.text

        if tag and dry_run:
            assert "Would tag" in caplog.text
        elif should_tag:
            assert "Tagging " in caplog.text
        else:
            assert "Would not tag" in caplog.text



"""Tests for the config.__init__ module."""

from dataclasses import dataclass
from typing import Optional

from bumpversion.scm import SCMInfo
from bumpversion.config import check_current_version


@dataclass
class MockConfig:
    """Just includes the current_version and scm_info attributes."""

    current_version: Optional[str]
    scm_info: SCMInfo


class TestCheckCurrentVersion:
    """
    Tests for the check_current_version function.

    - tag may not match serialization
    - tag may not match current version in config
    """

    def test_uses_tag_when_missing_current_version(self):
        """When the config does not have a current_version, the last tag is used."""
        # Assemble
        scm_info = SCMInfo()
        scm_info.current_version = "1.2.3"
        config = MockConfig(current_version=None, scm_info=scm_info)

        # Act
        result = check_current_version(config)

        # Assert
        assert result == "1.2.3"



"""Tests for the run_hooks function."""

from bumpversion import hooks
import pytest

from bumpversion.exceptions import HookError


def test_calls_each_hook(mocker):
    """It should call each hook passed to it."""
    # Assemble
    mock_logger = mocker.patch("bumpversion.hooks.logger")
    mock_run_command = mocker.patch("bumpversion.hooks.run_command")
    hooks_list = ["script1", "script2"]
    env = {"var": "value"}
    mock_run_command.return_value = mocker.MagicMock(stdout="output", stderr="error", returncode=0)

    # Act
    hooks.run_hooks(hooks_list, env)

    # Assert
    expected_info_calls = [
        mocker.call("Running 'script1'"),
        mocker.call("Running 'script2'"),
    ]
    mock_logger.info.assert_has_calls(expected_info_calls)
    expected_debug_calls = [
        mocker.call("Exited with 0"),
        mocker.call("output"),
        mocker.call("error"),
        mocker.call("Exited with 0"),
        mocker.call("output"),
        mocker.call("error"),
    ]
    mock_logger.debug.assert_has_calls(expected_debug_calls)
    mock_run_command.assert_any_call("script1", env)
    mock_run_command.assert_any_call("script2", env)


def test_raises_exception_if_hook_fails(mocker):
    """If a hook responds with an error, an exception should be raised."""
    # Assemble
    mock_logger = mocker.patch("bumpversion.hooks.logger")
    mock_run_command = mocker.patch("bumpversion.hooks.run_command")
    hooks_list = ["script1", "script2"]
    env = {"var": "value"}
    mock_run_command.return_value = mocker.MagicMock(stdout="output", stderr="error", returncode=1)

    # Act
    with pytest.raises(HookError):
        hooks.run_hooks(hooks_list, env)

    # Assert
    expected_info_calls = [mocker.call("Running 'script1'")]
    expected_warning_calls = [mocker.call("output"), mocker.call("error")]
    mock_logger.info.assert_has_calls(expected_info_calls)
    mock_logger.warning.assert_has_calls(expected_warning_calls)
    mock_run_command.assert_any_call("script1", env)


def test_does_not_call_each_hook_when_dry_run(mocker):
    """It should not call each hook passed to it when dry_run is True."""
    # Assemble
    mock_logger = mocker.patch("bumpversion.hooks.logger")
    mock_run_command = mocker.patch("bumpversion.hooks.run_command")
    hooks_list = ["script1", "script2"]
    env = {"var": "value"}

    # Act
    hooks.run_hooks(hooks_list, env, dry_run=True)

    # Assert
    expected_calls = [mocker.call("Would run 'script1'"), mocker.call("Would run 'script2'")]
    mock_logger.info.assert_has_calls(expected_calls)
    mock_run_command.assert_not_called()



import subprocess
from typing import Callable

import pytest
from pytest import param
from bumpversion.hooks import run_setup_hooks, run_pre_commit_hooks, run_post_commit_hooks
from tests.conftest import get_config_data, get_semver


def get_hook_env(*args, **kwargs) -> dict:
    """Mocked function for the environment setup"""
    return {}


def run_command(script: str, env: dict) -> subprocess.CompletedProcess:
    """Mocked function for command execution"""
    return subprocess.CompletedProcess(args=script, returncode=0)


CURRENT_VERSION = get_semver("1.0.0")
NEW_VERSION = get_semver("1.1.0")


class TestHookSuites:
    """Run each hook suite through the same set of tests."""

    suites = (
        param("setup", run_setup_hooks, (CURRENT_VERSION,), id="setup"),
        param(
            "pre_commit",
            run_pre_commit_hooks,
            (
                CURRENT_VERSION,
                NEW_VERSION,
            ),
            id="pre_commit",
        ),
        param(
            "post_commit",
            run_post_commit_hooks,
            (
                CURRENT_VERSION,
                NEW_VERSION,
            ),
            id="post_commit",
        ),
    )

    @pytest.mark.parametrize(["suite_name", "suite_func", "suite_args"], suites)
    def test_calls_each_hook(self, mocker, suite_name: str, suite_func: Callable, suite_args: tuple):
        """The suite hook runs each hook."""
        # Assemble
        env = {"var": "value"}
        mock_env = mocker.patch(f"bumpversion.hooks.get_{suite_name}_hook_env")
        mock_env.return_value = env
        mock_run_command = mocker.patch("bumpversion.hooks.run_command")
        mock_run_command.return_value = mocker.MagicMock(stdout="output", stderr="error", returncode=0)
        mock_logger = mocker.patch("bumpversion.hooks.logger")

        config, _, _ = get_config_data({"current_version": "1.0.0", f"{suite_name}_hooks": ["script1", "script2"]})

        # Act
        suite_func(config, *suite_args)

        # Assert
        expected_info_calls = [
            mocker.call(f"Running {suite_name} hooks:".replace("_", "-")),
            mocker.call("Running 'script1'"),
            mocker.call("Running 'script2'"),
        ]
        mock_logger.info.assert_has_calls(expected_info_calls)
        mock_env.assert_called_once_with(config, *suite_args)
        expected_run_command_calls = [
            mocker.call("script1", env),
            mocker.call("script2", env),
        ]
        mock_run_command.assert_has_calls(expected_run_command_calls)

    @pytest.mark.parametrize(["suite_name", "suite_func", "suite_args"], suites)
    def test_does_not_run_hooks_if_none_are_specified(
        self, mocker, suite_name: str, suite_func: Callable, suite_args: tuple
    ):
        """If no setup_hooks are defined, nothing is run."""
        # Assemble
        env = {"var": "value"}
        mock_env = mocker.patch(f"bumpversion.hooks.get_{suite_name}_hook_env")
        mock_env.return_value = env
        mock_run_command = mocker.patch("bumpversion.hooks.run_command")
        mock_run_command.return_value = mocker.MagicMock(stdout="output", stderr="error", returncode=0)
        mock_logger = mocker.patch("bumpversion.hooks.logger")

        config, _, _ = get_config_data({"current_version": "1.0.0", f"{suite_name}_hooks": []})

        # Act
        suite_func(config, *suite_args)

        # Asserts
        mock_logger.info.assert_called_once_with(f"No {suite_name} hooks defined".replace("_", "-"))
        mock_env.assert_called_once_with(config, *suite_args)
        assert mock_run_command.call_count == 0

    @pytest.mark.parametrize(["suite_name", "suite_func", "suite_args"], suites)
    def test_does_not_run_hooks_if_dry_run_is_true(
        self, mocker, suite_name: str, suite_func: Callable, suite_args: tuple
    ):
        """If dry_run is True, nothing is run."""
        # Assemble
        env = {"var": "value"}
        mock_env = mocker.patch(f"bumpversion.hooks.get_{suite_name}_hook_env")
        mock_env.return_value = env
        mock_run_command = mocker.patch("bumpversion.hooks.run_hooks")
        mock_logger = mocker.patch("bumpversion.hooks.logger")

        config, _, _ = get_config_data({"current_version": "1.0.0", f"{suite_name}_hooks": ["script1", "script2"]})

        # Act
        args = [*suite_args, True]
        suite_func(config, *args)

        # Asserts
        mock_logger.info.assert_called_once_with(f"Would run {suite_name} hooks:".replace("_", "-"))
        mock_env.assert_called_once_with(config, *suite_args)
        assert mock_run_command.call_count == 1



"""Tests for environment generation for hooks."""

import datetime
import os
import subprocess
from pathlib import Path

from bumpversion.hooks import (
    scm_env,
    PREFIX,
    base_env,
    version_env,
    new_version_env,
    get_setup_hook_env,
    get_pre_commit_hook_env,
)
from tests.conftest import inside_dir, get_config_data


def assert_os_environ_items_included(result_env: dict) -> None:
    """Assert that the OS environment variables are in the result."""
    for var, value in os.environ.items():
        assert var in result_env
        assert result_env[var] == value


def assert_scm_info_included(result_env: dict):
    """Assert the SCM information is included in the result."""
    assert f"{PREFIX}COMMIT_SHA" in result_env
    assert f"{PREFIX}DISTANCE_TO_LATEST_TAG" in result_env
    assert f"{PREFIX}IS_DIRTY" in result_env
    assert f"{PREFIX}BRANCH_NAME" in result_env
    assert f"{PREFIX}SHORT_BRANCH_NAME" in result_env
    assert f"{PREFIX}CURRENT_VERSION" in result_env
    assert f"{PREFIX}CURRENT_TAG" in result_env


def assert_current_version_info_included(result_env: dict):
    """Assert the current version information is included in the result."""
    assert f"{PREFIX}CURRENT_MAJOR" in result_env
    assert f"{PREFIX}CURRENT_MINOR" in result_env
    assert f"{PREFIX}CURRENT_PATCH" in result_env


def assert_new_version_info_included(result_env: dict):
    """Assert the new version information is included in the result."""
    assert f"{PREFIX}NEW_MAJOR" in result_env
    assert f"{PREFIX}NEW_MINOR" in result_env
    assert f"{PREFIX}NEW_PATCH" in result_env
    assert f"{PREFIX}NEW_VERSION" in result_env
    assert f"{PREFIX}NEW_VERSION_TAG" in result_env


def test_scm_env_returns_correct_info(git_repo: Path):
    """Should return information about the latest tag."""
    readme = git_repo.joinpath("readme.md")
    readme.touch()
    tag_prefix = "v"
    overrides = {"current_version": "0.1.0", "commit": True, "tag": True, "tag_name": f"{tag_prefix}{{new_version}}"}

    with inside_dir(git_repo):
        # Add a file and tag
        subprocess.run(["git", "add", "readme.md"])
        subprocess.run(["git", "commit", "-m", "first"])
        subprocess.run(["git", "tag", f"{tag_prefix}0.1.0"])
        conf, _, _ = get_config_data(overrides)

    result = scm_env(conf)
    assert result[f"{PREFIX}BRANCH_NAME"] == "master"
    assert len(result[f"{PREFIX}COMMIT_SHA"]) == 40
    assert result[f"{PREFIX}CURRENT_TAG"] == "v0.1.0"
    assert result[f"{PREFIX}CURRENT_VERSION"] == "0.1.0"
    assert result[f"{PREFIX}DISTANCE_TO_LATEST_TAG"] == "0"
    assert result[f"{PREFIX}IS_DIRTY"] == "False"
    assert result[f"{PREFIX}SHORT_BRANCH_NAME"] == "master"


class MockDatetime(datetime.datetime):
    @classmethod
    def now(cls, tz=None):
        return cls(2022, 2, 1, 17) if tz else cls(2022, 2, 1, 12)


class TestBaseEnv:
    """Tests for base_env function."""

    def test_includes_now_and_utcnow(self, mocker):
        """The output includes NOW and UTCNOW."""
        mocker.patch("datetime.datetime", new=MockDatetime)
        config, _, _ = get_config_data({"current_version": "0.1.0"})
        result_env = base_env(config)

        assert f"{PREFIX}NOW" in result_env
        assert f"{PREFIX}UTCNOW" in result_env
        assert result_env[f"{PREFIX}NOW"] == "2022-02-01T12:00:00"
        assert result_env[f"{PREFIX}UTCNOW"] == "2022-02-01T17:00:00"

    def test_includes_os_environ(self):
        """The output includes the current process' environment."""
        config, _, _ = get_config_data({"current_version": "0.1.0"})
        result_env = base_env(config)

        assert_os_environ_items_included(result_env)

    def test_includes_scm_info(self):
        """The output includes SCM information."""
        config, _, _ = get_config_data({"current_version": "0.1.0"})
        result_env = base_env(config)

        assert_scm_info_included(result_env)


def test_current_version_env_includes_correct_info():
    """The version_env for a version should include all its parts"""
    config, _, current_version = get_config_data(
        {"current_version": "0.1.0", "parse": r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)"}
    )
    result = version_env(current_version, "CURRENT_")

    assert result[f"{PREFIX}CURRENT_MAJOR"] == "0"
    assert result[f"{PREFIX}CURRENT_MINOR"] == "1"
    assert result[f"{PREFIX}CURRENT_PATCH"] == "0"


def test_new_version_env_includes_correct_info():
    """The new_version_env should return the serialized version and tag name."""

    config, _, current_version = get_config_data(
        {"current_version": "0.1.0", "parse": r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)"}
    )
    new_version = current_version.bump("minor")
    result = new_version_env(config, current_version, new_version)

    assert result[f"{PREFIX}NEW_VERSION"] == "0.2.0"
    assert result[f"{PREFIX}NEW_VERSION_TAG"] == "v0.2.0"


def test_get_setup_hook_env_includes_correct_info():
    """The setup hook environment should contain specific information."""
    config, _, current_version = get_config_data({"current_version": "0.1.0"})
    result_env = get_setup_hook_env(config, current_version)

    assert_os_environ_items_included(result_env)
    assert_scm_info_included(result_env)
    assert_current_version_info_included(result_env)


def test_get_pre_commit_hook_env_includes_correct_info():
    """The pre-commit hook environment should contain specific information."""
    config, _, current_version = get_config_data({"current_version": "0.1.0"})
    new_version = current_version.bump("minor")
    result_env = get_pre_commit_hook_env(config, current_version, new_version)

    assert_os_environ_items_included(result_env)
    assert_scm_info_included(result_env)
    assert_current_version_info_included(result_env)
    assert_new_version_info_included(result_env)



#!/usr/bin/env python
"""
Generate documentation stubs for the configured package.

Configuring this script:

    Change the `package_name` variable to the name of the package you want to generate

Output:

    This script will generate a `reference/api` directory containing documentation stubs for the package.
    It will also generate a `reference/api/SUMMARY.md` file that contains the navigation structure for the
    documentation.
"""

from pathlib import Path

import mkdocs_gen_files

package_name = "bumpversion"

nav = mkdocs_gen_files.Nav()
mod_symbol = '<code class="doc-symbol doc-symbol-nav doc-symbol-module"></code>'

src_root = Path(__file__).parent.parent
package_root = src_root / package_name

for path in sorted(package_root.rglob("*.py")):
    module_path = path.relative_to(src_root).with_suffix("")
    doc_path = path.relative_to(src_root).with_suffix(".md")
    full_doc_path = Path("reference/api", doc_path)

    parts = tuple(module_path.parts)
    if parts[-1] == "__init__":
        parts = parts[:-1]
        doc_path = doc_path.with_name("index.md")
        full_doc_path = full_doc_path.with_name("index.md")
    elif parts[-1].startswith("_"):
        continue

    nav_parts = [f"{mod_symbol} {part}" for part in parts]
    nav[tuple(nav_parts)] = doc_path.as_posix()

    with mkdocs_gen_files.open(full_doc_path, "w") as fd:
        ident = ".".join(parts)
        fd.write(f"::: {ident}")

    mkdocs_gen_files.set_edit_path(full_doc_path, path)

with mkdocs_gen_files.open("reference/api/nav.md", "w") as nav_file:
    nav_file.writelines(nav.build_literate_nav())
